/*
   Candle emulation.
   
   Changelog:
   	
   	2017-12-29: Initial release.
   	
*/

#include <math.h>
#include <stdlib.h>

#include "hal/led.h"
#include "hal/battery.h"
#include "hal/timer.h"

#include "settings/candle.h"

extern volatile int_fast8_t lowBatt;
extern const uint16_t redPow[4096];

static int_fast16_t in_range(int_fast16_t arg){
	
	if(arg > 4095){
		
		return 4095;
	}
	
	if(arg < 0){
		
		return 0;
	}
	
	return arg;
}


int main(void)
{
	
	int_fast16_t i;
	int_fast8_t j;
	int_fast16_t actual_bright;
	int_fast16_t nominal_bright;
	int8_t nominal_angle;
	int8_t actual_angle;
	int_fast8_t sign;
	int_fast16_t bright_addend[2];
	
	
	nominal_bright = CANDLE_MAX / 2;
	actual_bright = CANDLE_MAX / 2;
	nominal_angle = 0;
	actual_angle = 0;
	
	
	led_init();
	battery_init();
	timer_init();
	
	
	/* Something like a brownout detection on startup */
	
	while(!timer_ms(50)){
	}
	lowBatt = 0;
	
	
	
	
	
	while(1) {
		
		if(lowBatt){
			
			for(i = 0; i < 2001; i = i + 2){
				
				for(j = 0; j < 4; ++j){
					
					led_set(j, RED, i);
					led_set(j, GREEN, 2000 - i);
					led_set(j, BLUE, 0);
					
				}
			}
			
			timer_ms(0);
			while(!timer_ms(200)){
			}
			
			for(j = 0; j < 4; ++j){
					
				led_set(j, RED, 0);
				led_set(j, GREEN, 0);
				led_set(j, BLUE, 0);
			}
			
			while(!timer_ms(5000)){
			}
		}
		else{
		
			if(timer_ms(CANDLE_SPEED)){
				
				/* randomly change nominal brightness of the LEDs (Monte-Carlo inspired) */
			
				if (nominal_bright > rand() % CANDLE_MAX){
					
					if(rand() % 4095 < CANDLE_PROB){
					
						nominal_bright -= rand() % CANDLE_DOWN;
					}
				}
				else{
				
					nominal_bright += rand() % CANDLE_UP;			
				}
				
				nominal_bright = in_range(nominal_bright);
				
				
				/* randomly change wind's nominal angle */
				
				if(rand() % 4095 < CANDLE_ANGLE_PROB){
				
					nominal_angle += pow((rand() % 8191 - 4095) / 4095.0, CANDLE_ANGLE_EXP) * CANDLE_ANGLE_CHANGE;
				}
				
				
				/* smoothen nominal brightness and angle (-> actual brightness and angle) */
				
				if(nominal_bright - actual_bright < 0){
					actual_bright += 1 + (nominal_bright - actual_bright) / CANDLE_LOWPASS_DOWN;
				}
				else{
					actual_bright += -1 + (nominal_bright - actual_bright) / CANDLE_LOWPASS_UP;
				}
				
				if(actual_bright < nominal_bright){
					
					sign = (nominal_bright > 128 + actual_bright) ? -1 : 1;
				}
				else{
				
					sign = (actual_bright > 128 + nominal_bright) ? -1 : 1;
				}
				
				actual_angle += ((nominal_angle > actual_angle) ? 1 : -1) + sign * (nominal_angle - actual_angle) / CANDLE_LOWPASS_ANGLE;
				
				
				/* calculate candle's position (brightness manipulation) */
				
				bright_addend[0] = sin(actual_angle / 128.0 * 3.1415) * (4095.0 - actual_bright) / 4095.0 * CANDLE_POSITION_MAX; 
				bright_addend[1] = cos(actual_angle / 128.0 * 3.1415) * (4095.0 - actual_bright) / 4095.0 * CANDLE_POSITION_MAX;
				
				/* set LED values */
				
				for(i = 0; i < 2; ++i){
					
					/* 
					   LED i and LED i+2 are face to face. Add calculated addend (for
					   each axis (0 and 1)) to LED i and substract addend from LED i+2
					   in order to simulate a change of position.
					   
					   The red channel has a nonlinear characteristic in order
					   to make the light more red when brightness is low.
					*/
					
					led_set(i, 0, redPow[in_range(actual_bright - bright_addend[i])]);
					led_set(i, 1, in_range(actual_bright - bright_addend[i]));
					led_set(i, 2, in_range(actual_bright - bright_addend[i]));
					
					led_set(i + 2, 0, redPow[in_range(actual_bright + bright_addend[i])]);
					led_set(i + 2, 1, in_range(actual_bright + bright_addend[i]));
					led_set(i + 2, 2, in_range(actual_bright + bright_addend[i]));
				}
			}
		}		
	}

	return 0;
}
